# Mypy; for the `|` operator purpose
# Remove this __future__ import once the oldest supported Python is 3.10
from __future__ import annotations

from typing import Any, TYPE_CHECKING

from .api import jobs
from .backend_connection import BackendConnection
from .estimate_result import EstimateResult
from .job_result import JobResult

if TYPE_CHECKING:
    import datetime
    from pathlib import Path

# TODO this requires imports of actual quantum libraries for proper type
# checking.
CircuitT = Any


class DeqartClient:
    def __init__(
        self, api_token: str | None = None, config_location: str | Path | None = None
    ):
        self._backend_connection = BackendConnection(api_token, config_location)

    def search(
        self,
        run_status: str | None = None,
        created_later_than: str | datetime.datetime | None = None,
    ) -> list[JobResult]:
        """Search for jobs of the user
        :param run_status: if not None, run status of jobs to filter, can be one of "PENDING", "QUEUED", "RUNNING", "COMPLETED"
        :type run_status: str
        :param created_later_than: if not None, filter by latest datetime. Please add timezone for clarity, otherwise UTC will be assumed
        :type created_later_than: str or datetime.datetime
        :return: Metadata of jobs
        :rtype: list of JobResult
        """
        job_results = jobs.search_jobs(
            self._backend_connection, run_status, created_later_than
        )
        return [
            JobResult(dict(zip(job_results["column_names"], r)))
            for r in job_results["data"]
        ]

    def estimate(
        self,
        circuit: CircuitT,
    ) -> EstimateResult:
        """Estimate job runtime
        :param circuit: Quantum circuit
        :type circuit: Cirq, Qiskit, circuit
        :return: Estimate result class
        :rtype: EstimateResult
        """
        return EstimateResult(
            jobs.estimate_job_runtime(self._backend_connection, circuit)
        )

    def run(
        self,
        circuit: CircuitT,
        asynchronous: bool = False,
        debug_size: str | None = None,
    ) -> JobResult:
        """Run job on Deqart Platform
        :param circuit: Quantum circuit
        :type circuit: Cirq, Qiskit, circuit
        :param asynchronous: if set to False, wait for job completion before returning. If set to True, return immediately
        :type asynchronous: bool
        :return: JobResult metadata
        :rtype: JobResult
        """
        submitted_job = jobs.submit_job(self._backend_connection, circuit, debug_size)
        if not asynchronous:
            return self.wait(submitted_job["job_id"])
        return JobResult(submitted_job)

    def wait(self, job_id: str) -> JobResult:
        """Wait for job completion
        :param job_id: job_id that can be found as property of JobResult metadata
        :type job_id: str
        :return: JobResult metadata
        :rtype: JobResult
        """
        return JobResult(jobs.wait_for_job(self._backend_connection, job_id))



import base64
import io


def encode_circuit(qc):
    """
    Encode a quantum circuit object from various quantum libraries to a
    portable format.
    """
    # We use the string representation of the object class so that we don't
    # have to require users to install Qiskit, Cirq, and Braket all at once.
    circuit_type = str(type(qc))
    if circuit_type == "<class 'cirq.circuits.circuit.Circuit'>":
        import cirq

        # https://quantumai.google/cirq/build/interop
        serialized = cirq.to_json(qc)
        return {"circuit_type": "Cirq", "circuit": serialized}
    elif circuit_type == "<class 'qiskit.circuit.quantumcircuit.QuantumCircuit'>":
        from qiskit import qpy

        # https://qiskit.org/documentation/apidoc/qpy.html
        with io.BytesIO() as f:
            qpy.dump(qc, f)
            # Rewind back to the beginning of the "file".
            f.seek(0)
            serialized = f.read()
            # Convert to base64
            serialized = base64.b64encode(serialized)
            # Convert bytes to str
            serialized = serialized.decode("utf-8")
        return {"circuit_type": "Qiskit", "circuit": serialized}
    elif circuit_type == "<class 'braket.circuits.circuit.Circuit'>":
        from braket.circuits.serialization import IRType

        serialized = qc.to_ir(IRType.OPENQASM).source
        return {"circuit_type": "Braket", "circuit": serialized}
    else:
        raise Exception("Unsupported circuit type", qc)


def decode_circuit(data):
    """
    Decode the output of encode_circuit.
    """
    if "circuit_type" not in data:
        raise Exception("Wrong data structure. 'circuit_type' must be present.")
    if "circuit" not in data:
        raise Exception("Wrong data structure. 'circuit' must be present.")
    circuit_type = data["circuit_type"]
    if circuit_type == "Cirq":
        import cirq

        return cirq.read_json(json_text=data["circuit"])
    elif circuit_type == "Qiskit":
        from qiskit import qpy

        _bytes = str.encode(data["circuit"])
        _bytes = base64.b64decode(_bytes)
        with io.BytesIO(_bytes) as f:
            circuits = qpy.load(f)
            if len(circuits) != 1:
                raise Exception(
                    f"Unexpected number of circuits: {len(circuits)}. Must be 1."
                )
            return circuits[0]
    elif circuit_type == "Braket":
        raise Exception("Braket decoding is not yet supported.")
    else:
        raise Exception("Unsupported circuit type", circuit_type)



class EstimateResult:
    def __init__(self, data):
        self.device = data.get("device")
        self.estimate_ms = data.get("estimate_ms")
        self.num_qubits = data.get("num_qubits")
        self.circuit = data.get("qc")
        self.warning_message = data.get("warning_message")



import collections
import datetime
from io import BytesIO

import numpy as np
import requests
from dateutil.parser import parse


class JobResult:
    def __init__(self, data):
        #  "PENDING" | "QUEUED" |"RUNNING" | "COMPLETED"
        self.run_status = data.get("run_status")
        self.job_id = data.get("job_id")
        self.sv_link = data.get("sv_link")
        self.top_100_results = data.get("top_100_results")
        self.num_qubits = data.get("num_qubits")
        self.circuit = data.get("qc")
        self.tags = data.get("tags")

        self.created_on = data.get("created_on")

        self.queue_start = data.get("queue_start")
        if self.queue_start is not None:
            self.queue_start = parse(self.queue_start)

        self.run_start = data.get("run_start")
        if self.run_start is not None:
            self.run_start = parse(self.run_start)

        self.run_end = data.get("run_end")
        if self.run_end is not None:
            self.run_end = parse(self.run_end)

        if self.queue_start is not None and self.run_start is not None:
            self.queue_time_ms = int(
                (self.run_start - self.queue_start) / datetime.timedelta(milliseconds=1)
            )
        else:
            self.queue_time_ms = None

        if self.run_start is not None and self.run_end is not None:
            self.run_time_ms = round(
                (self.run_end - self.run_start) / datetime.timedelta(milliseconds=1)
            )
        else:
            self.run_time_ms = None

        Status = collections.namedtuple("Status", ["success", "error_message"])
        self.status = Status(data.get("success"), data.get("error_message"))

        self.cost = data.get("cost")
        if self.cost is not None:
            self.cost /= 100.0

    def get_statevector(self):
        if self.sv_link is not None:
            data = requests.get(self.sv_link, timeout=100).content
            return np.load(BytesIO(data))
        return None

    def __str__(self):
        return f"Job ID: {self.job_id}, run status: {self.run_status}, queue time (ms): {self.queue_time_ms}, run time (ms): {self.run_time_ms}, cost ($): {self.cost}, num qubits: {self.num_qubits}, status: {self.status}"



from setuptools import find_packages, setup

with open("requirements.txt") as f:
    requirements = f.read()
requirements = requirements.splitlines()

with open("README.md", encoding="utf-8") as f:
    readme = f.read()
readme = "\n".join(readme.split("\n")[2:])

packages = find_packages()

with open("deqart/version.py") as f:
    Version = f.read()

Version = Version.rstrip()
Version = Version[15:-1]

setup(
    name="deqart",
    version=Version,
    description="Python SDK to Deqart platform",
    license="Apache 2.0",
    author="BlueQubit",
    author_email="hovnatan@bluequbit.io",
    url="https://github.com/BlueQubitDev/deqart-python-sdk",
    long_description=readme,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    setup_requires=["wheel"],
    packages=find_packages(exclude=("tests",)),
    entry_points={"console_scripts": ["deqart = deqart.__main__:main"]},
    python_requires=">=3.6",
)



import logging

logger = logging.getLogger("deqart-python-sdk")


class DeqartBaseException(Exception):
    def __init__(
        self, status_code="Unknown Status Code", message="Unknown Error Message"
    ):
        self.status_code = status_code
        self.message = message
        super().__init__(status_code, message)


class DeqartUnauthorizedException(DeqartBaseException):
    def __init__(self):
        super().__init__(
            401,
            "Wrong API key. You can find your API key once you log in to https://www.deqart.com",
        )
